#!/usr/local/bin/perl

# Name: Selena Sol's Mailing List Manager
#
# Version: 3.0
#
# Last Modified: 01-22-97
#
# Copyright Info: This application was written by Selena Sol
# (selena@eff.org, http://www.eff.org/~erict) having been inspired by
# countless other Perl authors.  Feel free to copy, cite, reference,
# sample, borrow, resell or plagiarize the contents.  However, if you
# don't mind, please let me know where it goes so that I can at least
# watch and take part in the development of the memes. Information wants
# to be free, support public domain freware.  Donations are appreciated
# and will be spent on further upgrades and other public domain scripts.
#
# Finally, PLEASE SEND WORKING URL's to selena@eff.org.  I maintain a list
# of implementations around the world.

###########################################################################
#                       Print http Header.                                #
###########################################################################

                # The script begins by telling the Perl interpreter that
                # it should continuously flush its buffer so that text
                # from this script is sent directly to the Web Browser.
                # We do this to streamline debugging and make sure that
                # the script operates with the flow we want it to. 

$| = 1;

		# Next, the HTTP header is sent out so that the browser
		# won't time us out if we take a long time to process the
		# request.

print "Content-type: text/html\n\n";

##########################################################################
#                      Load Libraries.    				 #
##########################################################################

		# Next the script loads a few supporting libraries. 
		# cgi-lib.pl uses the subroutine ReadParse to read any
		# incoming data. mail-lib.pl is used to email
		# confirmations as well as perform the mass mailings
		# themselves. auth-lib.pl is used to perform
		# administrative authentication for mass mailings.

&require_supporting_libraries (__FILE__, __LINE__,
                              "./Library/cgi-lib.pl",
			      "./Library/mail-lib.pl",
			      "./Library/auth-lib.pl");

                # Next, the ReadParse subroutine in cgi-lib.pl is used to
                # read the incoming form data. However, the subroutine is
                # sent "form_data" as a parameter so that the associative
                # array of form keys/values comes back with a descriptinve
                # name rather than just %in.

&ReadParse(*form_data);

		# Then the script determines which mailing list it is
		# being asked to manipulate by accessing an incoming form
		# variable called "list".  List therefore may be
		# passed to this script by way of a hidden form variable
		# such as in the following example in which you want the
		# CGI to utilize a setup file called auto.setup.
		# 
		# <INPUT TYPE = "hidden" NAME = "list" VALUE = "auto">
		#
		# or as URL encoded information...
		# 
		# http://www.you.com/cgi/mailing_list.cgi?setup_file=auto
		#
		# Thus, this one script could handle many different
		# mailing lists if required.
		# 
		# default.setup is the default setup file, however.
		#
		# The setup files are used to provide server specific
		# information related to this script.

  if ($form_data{'setup_file'} ne "")
    {
    $setup_file = "$form_data{'setup_file'}.setup";
    }
  else
    {
    $setup_file = "default.setup";
    }

&require_supporting_libraries (__FILE__, __LINE__,
                              "./Setup_files/$setup_file");

###########################################################################
#                  Generate User Mailing List Form                        #
###########################################################################

		# The first duty of this script is to provide users with
		# a form so that they can add their address to the
		# database.  However, we need to make sure that that is
		# what they are asking for, because this script is also
		# used for other purposes, such as the administrative
		# functions.
		#
		# We'll do this by asking two questions.  
		# 
		# Firstly, is the incoming data GET or POST data.  If it
		# is POST data, we know that we have already shown the
		# user this page and they are POSTING their response back
		# to us (since we use the POST method for form input).  If
		# they just clicked on a hyperlink to get this script,
		# then thier request will be GET.  
		#
		# Note: If you need to change this logic, you could
		# always add a URL encoded string to references to the
		# script like 
		#
		# http://www.yourdomain.com/mailing_list.cgi?view=first_form
		#
		# Then, of course, you could change the if test to test
		# whether or not that particular form input variable has a
		# value...I just prefer to use the GET vrs POST
		# distinction.
		# 
		# And (&&), if they were asking for the administrator's
		# page, we should not give them the frontpage form...so if
		# action=something, we want to skip this if test as well. 
		# "action" is another special keyword that can be passed
		# to this script via hidden form tag or URL encoded data
		# just as "list" was above.

		# If both conditions are true, we can print out the add
		# form using the user_form subroutine in the setup file to
		# display.

if ($ENV{'REQUEST_METHOD'} eq "GET" && $form_data{'action'} eq "")
  {
  &user_form;
  exit;
  }

#######################################################################   
#                       Check Required Fields                         #   
#######################################################################
                
                # Now that the script has loaded all of the supporting
                # libraries it is time to start processing the incoming
                # form data.
                #
                # The first thing that the script does is to make sure
                # that the user has filled in data for every field
                # that the administrator has defined as "required".  The
                # list of required fields are defined in the setup file
                # array @required_variables for each separate form to be
                # processed.
                #
                # The script simply goes through each of the fields and
                # checks to see if it has a value coming in as form data
                # (as stored in the form_data associative aray by  
                # cgi-lib.pl).  If the field was left blank, the script
                # accesses the subroutine required_fields_error_message in
                # the setup file which sends the user a note regarding the
                # error and then exits.
		#
		# Also, we will bypass the required fields check if we are
		# generating the administrative screens because we will
		# assume that there are no required fields for that.
                         
foreach $variable (@required_variables)
  {
  if ($form_data{$variable} eq "" && $form_data{'action'} eq "" &&
      $form_data{'mail_to_list'} eq "")
    {
    &required_fields_error_message;
    exit;       
    }
  }

###########################################################################
#                  Add New Entry to Mailing List                          #
###########################################################################

		# If the user just filled out her information on the add
		# form generated above, she will have hit the submit
		# button which was named "new_entry" so that
		# $form_data{'new_entry'} will equal "Send me Your
		# Information" and we'll know that we are being asked to
		# add a new name to the list.
		#
		# Note: When you are designing your own forms to work with
		# this script, you are welcome to change the VALUE
		# parameter of the submit button to anything you'd like,
		# but the NAME parameter must remain equal to "new_entry"
		# because that value is used here.

if ($form_data{'new_entry'} ne "")
  {

		# Once the script has determined that it needs to add a
		# new entry(s), it will look for the special incoming form
		# variable called "lists".
		#
		# This special variable will be used to determine which
		# datafiles the new entry should be added to. Afterall
		# the user may wish to add themselves to one or several
		# lists at once.
		#
		# Thus, you might have one of several FORM input methods
		# to define "lists" within your HTML.  Here are a couple
		#
		# <INPUT TYPE = "text" NAME = "lists">
		#
		# This is not so great because the user may make a typo
		# when she selects which datafile to be added to.
		# A preferrable method would be to use a hidden input
		# field such as the following:
		#
		# <INPUT TYPE = "hidden" NAME = "lists" VALUE = "auto">
		#
		# Thus, the user would automatically have her data
		# submitted to auto.list.  UNfortunately, you may want to
		# allow the user to choose from several lists with a
		# SELECT box such as the following:
		#
		# <SELECT NAME = "lists">	
		# <OPTION>auto
		# <OPTION>plane
		# </SELECT>
		#
		# Finally, you may wish to allow the user to sign up to
		# multiple lists at once with the following HTML:
		#
                # <SELECT NAME = "lists" MULTIPLE> 
                # <OPTION>auto  
                # <OPTION>plane 
                # </SELECT>
		# 
		# The problem with all of these options is that this
		# script needs to be able to handle the dicersity.  This
		# problem is solved by cgi-lib.pl and its SplitParam
		# subroutine.  This subroutine will break up the incoming
		# form data on MULTIPLE selects into an array of elements
		# as well as process single requests, dumping them as the
		# only element in that same array.
		#
		# In short, the following line will collect all of the
		# lists that the user selected (whether it is one list or
		# multiple lists) and dump them into the @lists array for
		# further processing.

  @lists = &SplitParam($form_data{'lists'});

		# There is one more special value.  Within your HTML, you
		# are allowed to use a special value of "all".  For
		# example, you might have the following HTML code:
		# 
                # <SELECT NAME = "lists">
                # <OPTION>auto  
                # <OPTION>plane
		# <OPTION>boat
		# <OPTION>all
                # </SELECT>
		#
		# If the client selects "all" then they should be added to
		# every list available in one shot.  However, the script
		# needs to know which lists are available for addition.
		# This info is provided in the setup file as @main_list.
		# If the user selects "all", the script will add them to
		# every list contained in @main_list.  Thus, if you add
		# lists to the HTML code, make suer to add them to
		# @main_list as well.  :)

  if ($form_data{'lists'} eq "all")
    {
    @lists = @main_list;
    }

		# So, now that the script has the array of lists to add
		# the user to, it is time to do so.  Thus, for every list
		# in the array, the script opens up the database file for
		# that list, but opens it so that we can add another
		# entry to ther end (>>) and adds them to the database
		# file.

  foreach $list (@lists)
    {

		# The script will create a pipe delimited database row out
		# of the fields that were defined in the setup file as
		# coming in from the form (@form_fields).  
		#
		# However, the script will change all occurances of the
		# pipe character (|) to ~p~ (because the pipe character is
		# needed to delimit the data) and all occurances of the
		# newline character (\n) to ~nl~ because the newline
		# character is used to delimit database rows.

    foreach $field (@form_fields)
      {
      $form_data{$field} =~ s/\n/~nl~/g;
      $form_data{$field} =~ s/\|/~p~/g;
      $database_row .= "$form_data{$field}\|";
      }

		# Once all the user defined fields have been added to the
		# new database row, the script will chop off the final
		# pipe character because we don't need one for the last
		# database field and will add the new entry to the
		# database file.  However, before actually  
                # modifying the database file, the script makes sure to
                # use the lockfile routines discussed later in this script
                # to assure that only one instance of the script can 
                # modify the file at one time.  Afterall, it would not do
                # for two instances of the script to modify the file at 
                # one time.  At best, one of the forms would not be
                # processed correctly.  At worst, the data file might  
                # become corrupt.  The lockfile routines prevent more than
                # one instance of the script from accessing the file at
                # one time.  Then the database file will be closed
		# up and the database row variable will be cleared so that
		# if the client is submitting to multiple lists, the
		# script will be able to address them with a clean
		# variable.  Notice that 

    &get_file_lock ("$database_directory_path/$list.lock");
    open (DATABASE, ">>$database_directory_path/$list.list") ||
	&file_open_error("$list.list", "Add New Entry to Mailing List",
			 __FILE__, __LINE__);
    chop $database_row;
    print DATABASE "$database_row\n";
    close (DATABASE);
    &release_file_lock ("$database_directory_path/$list.lock"); 
    $database_row = "";
    }

#######################################################################
#                       Respond to the Client                         # 
#######################################################################   
                
                # Finally, the script sends a response to the client.
                # There are two types of responses possible.  Firstly, the
                # client should receive and HTML thank you response.  This
                # is done using the thank_you subroutine defined in the
		# setup file so that you can easily customize this HTML
		# code.   
                
  &thank_you;

                # The second type of response is an email response.
                # Perhaps you want to send the user a separate email
                # notification letting them know that their info was sent
                # in.  In this case, provided that you had the
                # client_email filed, we will simply create an email
                # message just as we did for the administartor and mail
                # it.
		#
		# Note: This routine is only valid if 1) you have set an
		# HTML input field equal to NAME = "client_email" and 2)
		# if the user submits a valid email address for that field.
                          
  if ($should_i_send_user_email eq "yes")
    {
    if ($form_data{'client_email'} ne "")
      {
      $email_body = "This email is simply a notification that you have
      been added to our mailing lists. As a reminder, you submitted the
      following data at $ENV{'HTTP_REFERER'} on: ";
      $email_body .= &get_date;
      $email_body .="\n\n";
  
      foreach $variable (@form_variables)
        {
        $email_body .= &format_text_field($form_variable_name_map{$variable});
        $email_body .= "$form_data{$variable}\n";
        }
        &send_mail("$email_of_sender","$form_data{'client_email'}",
                 "$email_subject", "$email_body");
      }
    }

  exit;
  }

###########################################################################
#                    Authenticate Administrator                           #
###########################################################################

		# If we have received an incoming variable named action
		# with a value of "admin" AND we don't have a variable
		# mail_to_list, then that means that  we should validate
		# the admin.  (in this routine we'll generate an 
		# interface for the admin with a submit button called
		# mail_to_list) :)

  if ($form_data{'action'} eq "admin" && $form_data{'mail_to_list'} eq "")
    {

		# Now, let's use the subroutine &GetSessionInfo in
		# auth-lib.pl to check the user against the user
		# database.  We pass three parametrs to the  subroutine
		# $session_file, $this_script_location, and *form_data.  
		# *form_data is equal to the associative array we got
		# from cgi-lib.pl.  $this_script_location was defined in
		# the setup file.  But $session_file is going to be blank
		# because we have not made one yet. Thus,. the 
		# authentication scripts will know to validate as a new
		# user... If the user enters a valid userid and password,
		# it will create a session file for them and we'll pass
		# the session file info along as a hidden form variable
		# at the end of this routine.

    ($session_file, $session_username, $session_group,$session_first_name,
     $session_last_name, $session_email) =
     &GetSessionInfo($session_file, $this_script_url, *form_data);

		# Okay, now output a form so that the admin can prepare
		# the email for mass mailing.
    &admin_form;
    exit;
    }

###########################################################################
#                     Do a Mass Mailing                                   #
###########################################################################

		# Okay, if we have gotten this far, it means that we have
		# been asked by a validated admin to do a mass mailing
		# when she hit the mail_to_list' submit button generated
		# in the routine above....so let's go ahead.

  if ($form_data{'mail_to_list'} ne "")
    {

		# However, let's run them through a security check once
		# more by sending their session_file through another
		# security check to make sure they are actually who they
		# say they are.

    ($session_file, $session_username, $session_group, $session_first_name,
     $session_last_name, $session_email) =
     &GetSessionInfo($form_data{'session_file'}, $this_script_location, 
     *form_data);

		# Also, make sure that the admin entered both a body AND
		# a subject...if not, have them try again. (|| means or).

    if ($form_data{'subject'} eq "" || $form_data{'comments'} eq "")
      {
      &admin_error;
      exit;
      }

		# Now open up our list of mailing address and begin going
		# through them one line at a time (while (<DATABASE>))

    $database = "$database_directory_path$form_data{'database'}";
    open (DATABASE, "$database") ||
         &file_open_error("$database", "Mass Mailing",
                          _FILE__, __LINE__); 

    while (<DATABASE>)
      {  

		# Each line will have someones email followed by a pipe
		# (|) followed by their name.  We only really care about
		# their email, so we'll split each line into name and
		# email using the pipe as our delimiter.

      ($recipient_name, $recipient_email, @junk) = split (/\|/, $_);

		# Now use the send_mail routine in mail-lib.pl to send
		# each person the email.  The routine takes 4
		# parameters...who the mail is from, who it is to, the
		# subject and the body)

      if ($form_data{'list'} ne "demo.setup")
        {
        &send_mail("$email_from","$recipient_email",
                   "$form_data{'subject'}", 
	  	   "$form_data{'comments'}");
        }

		# Now print out a little list of successful emails for
		# the admin and close up the database.

      print "Mail Sent to: $recipient_email<BR>";
      }
    close (DATABASE);
    exit;
    }

####################################################################### 
#                       Require Supporting Libraries.                 #
#######################################################################   
                
                # require_supporting_libraries is used to read in some of
                # the supporting files that this script will take
                # advantage of. 
                #
                # require_supporting_libraries takes a list of arguments
                # beginning with the current filename, the current line
                # number and continuing with the list of files which must
                # be required using the following syntax:
                #
                # &require_supporting_libraries (__FILE__, __LINE__, 
                #                               "file1", "file2",
                #                               "file3"...);

                #
                # Note: __FILE__ and __LINE__ are special Perl variables
                # which contain the current filename and line number   
                # respectively.  We'll continually use these two variables
                # throughout the rest of this script in order to generate
                # useful error messages.
                
sub require_supporting_libraries
  {
                
                # The incoming file and line arguments are split into  
                # the local variables $file and $line while the file list
                # is assigned to the local list array @require_files.
                #
                # $require_file which will just be a temporary holder
                # variable for our foreach processing is also defined as a
                # local variable.
                 
  local ($file, $line, @require_files) = @_;
  local ($require_file);
                
                # Next, the script checks to see if every file in the
                # @require_files list array exists (-e) and is readable by
                # it (-r). If so, the script goes ahead and requires it.

  foreach $require_file (@require_files)
    {           
    if (-e "$require_file" && -r "$require_file")
      {
      require "$require_file";
      }
                
                # If not, the scripts sends back an error message that
                # will help the admin isolate the problem with the script.
                 
    else
      {
      print "Content-type: text/html\n\n";
      print "I am sorry but I was unable to require $require_file at line
            $line in $file.  Would you please make sure that you have the 
            path correct and that the permissions are set so that I have
            read access?  Thank you.";
      exit;
      }         
    } # End of foreach $require_file (@require_files)
  } # End of sub require_supporting_libraries

#######################################################################
#                    file_open_error Subroutine                       #
#######################################################################

                # If there is a problem opening a file or a directory, it
                # is useful for the script to output some information
                # pertaining to what problem has occurred.  This
                # subroutine is used to generate those error messages.
                #
                # file_open_error takes four arguments: the file or
                # directory which failed, the section in the code in which
                # the call was made, the current file name and
                # line number, and is called with the following syntax:
                #
                # &file_open_error("file.name", "ROUTINE", __FILE__,
                #                  __LINE__);

sub file_open_error
  {

                # The subroutine simply uses the update_error_log
                # subroutine discussed later to modify the error log and
                # then uses CgiDie in cgi-lib.pl to gracefully exit the
                # application with a useful debugging error message sent
                # to the browser window.

  local ($bad_file, $script_section, $this_file, $line_number) = @_;
  print "Content-type: text/html\n\n";
  &CgiDie ("I am sorry, but I was not able to access $bad_file in the
        $script_section routine of $this_file at line number $line_number.
        Would you please make sure the path is correctly defined in
        web_store.setup and that the permissions are correct.")
  }

#######################################################################
#                            get_file_lock                            #
#######################################################################
   
                # get_file_lock is a subroutine used to create a lockfile.
                # Lockfiles are used to make sure that no more than one
                # instance of the script can modify a file at one time.  A
                # lock file is vital to the integrity of your data.
                # Imagine what would happen if two or three people
                # were using the same script to modify a shared file (like
                # the error log) and each accessed the file at the same
                # time.  At best, the data entered by some of the users
                # would be lost.  Worse, the conflicting demands could
                # possibly result in the corruption of the file.
                #
                # Thus, it is crucial to provide a way to monitor and
                # control access to the file.  This is the goal of the
                # lock file routines.  When an instance of this script
                # tries to  access a shared file, it must first check for
                # the existence of a lock file by using the file lock
                # checks in get_file_lock.
                #
                # If get_file_lock determines that there is an existing
                # lock file, it instructs the instance that called it to  
                # wait until the lock file disappears.  The script then 
                # waits and checks back after some time interval.  If the 
                # lock file still remains, it continues to wait until some
                # point at which the admin has given it permissios to just
                # overwrite the file because some other error must have
                # occurred.
                #
                # If, on the other hand, the lock file has dissappeared,
                # the script asks get_file_lock to create a new lock file
                # and then goes ahead and edits the file.
                #
                # The subroutine takes one argumnet, the name to use for
                # the lock file and is called with the following syntax:
                #
                # &get_file_lock("file.name");

sub get_file_lock
  {
  local ($lock_file) = @_;
  local ($endtime);
  $endtime = 20;
  $endtime = time + $endtime;

                # We set endtime to wait 20 seconds.  If the lockfile has
                # not been removed by then, there must be some other
                # problem with the file system.  Perhaps an instance of
                # the script crashed and never could delete the lock file.

  while (-e $lock_file && time < $endtime)
    {
    sleep(1);
    }

  open(LOCK_FILE, ">$lock_file") || &file_open_error ("$lock_file",
                                                      "Lock File Routine",
                                                      __FILE__, __LINE__);

                # Note: If flock is available on your system, feel free to
                # use it.  flock is an even safer method of locking your
                # file because it locks it at the system level.  The above
                # routine is "pretty good" and it will server for most
                # systems.  But if youare lucky enough to have a server
                # with flock routines built in, go ahead and uncomment
                # the next line and comment the one above.

# flock(LOCK_FILE, 2); # 2 exclusively locks the file

  }
                 
####################################################################### 
#                            release_file_lock                        # 
#######################################################################
                
                # release_file_lock is the partner of get_file_lock.  When
                # an instance of this script is done using the file it
                # needs to manipulate, it calls release_file_lock to
                # delete the lock file that it put in place so that other
                # instances of the script can get to the shared file.  It
                # takes one argument, the name of the lock file, and is
                # called with the following syntax:
                #
                # &release_file_lock("file.name");
                
sub release_file_lock
  {
  local ($lock_file) = @_;
  
# flock(LOCK_FILE, 8); # 8 unlocks the file
    
                # As we mentioned in the discussion of get_file_lock,
                # flock is a superior file locking system.  If your system
                # has it, go ahead and use it instead of the hand rolled
                # version here.  Uncomment the above line and comment the 
                # two that follow.
                
  close(LOCK_FILE);
  unlink($lock_file);
  }               
